from __future__ import annotations

__all__ = ['Slide']

import os
from bisect import bisect_right
from pathlib import Path
from typing import final
from warnings import warn

import cv2
import numpy as np
from glow import memoize, shared_call, weak_memoize

from bipl import env
from bipl.ops import normalize_loc

from ._openslide import Openslide
from ._slide_bases import REGISTRY, Driver, Item, Lod
from ._tiff import Tiff

# TODO: inside Slide.open import ._slide.registry,
# TODO: and in ._slide.registry do registration and DLL loading
# TODO: to make Slide export not require DLL presence

try:
    from ._gdal import Gdal
except ImportError:
    if not os.getenv('_BIPL_GDAL_NO_WARN'):
        msg = 'No GDAL is available. Please '
        if os.name == 'nt':
            msg += ('acquire it manually from '
                    'https://github.com/cgohlke/geospatial-wheels/releases ')
        else:
            msg += ('install libgdal via your system package manager, '
                    'and run "pip install gdal==`gdal-config --version`"')
        warn(msg, stacklevel=1)
        os.environ['_BIPL_GDAL_NO_WARN'] = '1'
    Gdal = None  # type: ignore[assignment,misc]

_drv: type[Driver] | None
for _drv, _regex in [  # LIFO, last driver takes priority
    (Gdal, r'^.*\.(tif|tiff)$'),
    (Openslide, r'^.*\.(bif|mrxs|ndpi|scn|svs|svsslide|tif|tiff|vms|vmu)$'),
    (Tiff, r'^.*\.(svs|tif|tiff)$'),
    (Gdal, r'^(/vsicurl.*|(http|https)://.*)$'),
]:
    if _drv is not None and _drv.__name__.lower() in env.BIPL_DRIVERS:
        _drv.register(_regex)


@shared_call  # merge duplicate calls
@weak_memoize  # reuse result if it's already exist, but used by someone else
@memoize(capacity=env.BIPL_CACHE, policy='lru')  # keep LRU for unused results
def _cached_open(path: str) -> Slide:
    last_exc = BaseException()
    matches = (tp for pat, tps_ in REGISTRY.items() if pat.match(path)
               for tp in tps_)
    tps = [*dict.fromkeys(matches)]
    if tps:
        for tp in reversed(tps):  # Loop over types to find non-failing
            try:
                return Slide(path, tp)
            except (ValueError, TypeError) as exc:
                last_exc = exc
        raise last_exc from None
    raise ValueError(f'Unknown file format {path}')


def _fit_to(image: np.ndarray, dsize: tuple[int, ...]):
    if image.shape[:2] == dsize:
        return image
    return cv2.resize(image, dsize[::-1], interpolation=cv2.INTER_AREA)


@final
class Slide:
    """Usage:
    ```
    slide = Slide.open('test.svs')
    shape: tuple[int, ...] = slide.shape
    scales: tuple[int, ...] = slide.scales

    # Get numpy.ndarray
    image: np.ndarray = slide[:2048, :2048]
    ```
    """
    # TODO: check if memory leak
    # TODO: add __enter__/__exit__/close to make memory management explicit
    # TODO: call .close in finalizer
    path: str
    shape: tuple[int, ...]
    spacing: float | None
    pools: tuple[int, ...]
    lods: tuple[Lod, ...]
    extras: dict[str, Item]

    def __init__(self, path: str, driver_cls: type[Driver]):
        self.path = path
        driver = driver_cls(path)

        num_items = len(driver)
        if num_items == 0:
            raise ValueError('Empty file')

        item_0, *items = (driver[idx] for idx in range(num_items))
        if not isinstance(item_0, Lod):
            raise TypeError('First pyramid layer is not tiled')

        lods: dict[int, Lod] = {1: item_0}
        self.extras: dict[str, Item] = {}
        for item in items:
            if isinstance(item, Lod):
                pool = round(item_0.shape[0] / item.shape[0])
                lods[pool] = item
            elif key := item.key:
                self.extras[key] = item

        self.pools = *sorted(lods.keys()),
        # TODO: create virtual lods if pools are too distant (ProxyLod?)
        self.lods = *(lods[pool] for pool in self.pools),
        self.shape = self.lods[0].shape
        self.spacing = self.lods[0].spacing
        self.bbox = driver.bbox

        self.extras |= driver.named_items()
        self.driver = driver_cls.__name__

    def __reduce__(self) -> tuple:
        return Slide.open, (self.path, )

    def __repr__(self) -> str:
        line = f"'{self.path}', shape={self.shape}, pools={self.pools}"
        if self.spacing:
            line += f', spacing={self.spacing:.4f}'
        line += f', driver={self.driver}'
        return f'{type(self).__name__}({line})'

    def best_lod_for(self, pool: float) -> tuple[int, Lod]:
        """Gives the most detailed LOD below `pool`"""
        idx = max(bisect_right(self.pools, pool) - 1, 0)
        return self.pools[idx], self.lods[idx]

    def pool(self, zoom: float, *, eps: float = 0.01) -> Lod:
        """Use like `slide.pool(4)[y0:y1, x0:x1]` call"""
        p, lod = self.best_lod_for(zoom * max(1, 1 + eps))
        if p == zoom:
            return lod
        return lod.rescale(p / zoom)

    def __getitem__(self, key: slice | tuple[slice, ...]) -> np.ndarray:
        """Retrieves tile"""
        # TODO: Ignore step, always redirect to self.lods[0].__getitem__
        y_loc, x_loc, c_loc = normalize_loc(key, self.shape)

        step0, step1 = y_loc.step, x_loc.step
        if step0 != step1:
            raise ValueError('slice steps should be the same for each axis')
        if step0 <= 0:
            raise ValueError('slice steps should be positive')

        pool, lod = self.best_lod_for(step0)
        yx_loc = *(slice(s.start // pool, s.stop // pool)
                   for s in (y_loc, x_loc)),
        image = lod.crop(*yx_loc)

        ratio = pool / step0
        dsize = *(round(ratio * s) for s in image.shape[:2]),
        return _fit_to(image, dsize)[:, :, c_loc]

    def at(self,
           z0_yx_offset: tuple[int, ...],
           dsize: int | tuple[int, ...],
           scale: float = 1) -> np.ndarray:
        """Read square region starting with offset"""
        dsize = dsize if isinstance(dsize, tuple) else (dsize, dsize)
        if len(dsize) != 2:
            raise ValueError(f'dsize should be 2-tuple or int. Got {dsize}')

        pool, lod = self.best_lod_for(scale)
        loc = *(slice(int(c) // pool,
                      int(c + d * scale) // pool)
                for c, d in zip(z0_yx_offset, dsize)),
        image = lod.crop(*loc)
        return _fit_to(image, dsize)

    def extra(self, name: str) -> np.ndarray | None:
        if item := self.extras.get(name):
            return np.array(item, copy=False, order='C')
        return None

    def thumbnail(self) -> np.ndarray:
        item = self.extras.get('thumbnail')
        if not item:
            item = self.lods[-1]
        return np.asarray(item)

    @classmethod
    def open(cls, anypath: Path | str) -> Slide:
        """Open multi-scale image."""
        if isinstance(anypath, Path):
            anypath = Path(anypath).resolve().absolute().as_posix()
        return _cached_open(anypath)



from __future__ import annotations

__all__ = ['Driver', 'Item', 'Lod', 'REGISTRY']

import re
from dataclasses import dataclass
from typing import final

import cv2
import numpy as np

from bipl.ops import normalize_loc

REGISTRY: dict[re.Pattern, list[type[Driver]]] = {}


@dataclass(frozen=True)
class Item:
    shape: tuple[int, ...]

    @property
    def key(self) -> str | None:
        raise NotImplementedError

    def __array__(self) -> np.ndarray:
        raise NotImplementedError


@dataclass(frozen=True)
class Lod(Item):
    spacing: float | None

    @final
    @property
    def key(self) -> None:
        return None

    def crop(self, *loc: slice) -> np.ndarray:
        """Reads crop of LOD. Overridable"""
        raise NotImplementedError

    @final
    def __getitem__(self, key: slice | tuple[slice, ...]) -> np.ndarray:
        """Reads crop of LOD"""
        y_loc, x_loc, c_loc = normalize_loc(key, self.shape)
        if not y_loc.step == x_loc.step == 1:
            raise ValueError('Y/X slice steps should be 1 for now, '
                             f'got {y_loc.step} and {x_loc.step}')
        return self.crop(y_loc, x_loc)[:, :, c_loc]

    @final
    def __array__(self) -> np.ndarray:
        """Reads whole LOD in single call"""
        return self[:, :]

    def rescale(self, scale: float) -> Lod:
        if scale == 1:
            return self
        h, w, c = self.shape
        return ProxyLod(
            # TODO: round/ceil/floor ?
            (round(h * scale), round(w * scale), c),
            (self.spacing / scale if self.spacing else None),
            scale,
            self.base if isinstance(self, ProxyLod) else self,
        )

    def _unpack_loc(
        self,
        *slices: slice,
    ) -> tuple[np.ndarray, np.ndarray, list[int]]:
        box = np.array([(s.start, s.stop) for s in slices])
        valid_box = box.T.clip([0, 0], self.shape[:2]).T  # (2, lo-hi)
        shape = (box[:, 1] - box[:, 0]).tolist()
        return box, valid_box, shape

    def _expand(self, rgb: np.ndarray, valid_box: np.ndarray, box: np.ndarray,
                bg_color: np.ndarray) -> np.ndarray:
        offsets = np.abs(valid_box - box)
        if offsets.any():
            tp, bm, lt, rt = offsets.ravel().tolist()
            rgb = cv2.copyMakeBorder(rgb, tp, bm, lt, rt, cv2.BORDER_CONSTANT,
                                     None, bg_color.tolist())
        return np.ascontiguousarray(rgb)


@dataclass(frozen=True)
class ProxyLod(Lod):
    scale: float
    base: Lod

    def crop(self, *loc: slice) -> np.ndarray:
        src_loc = *[
            # TODO: round/ceil/floor ?
            slice(round(s.start / self.scale), round(s.stop / self.scale))
            for s in loc
        ],
        # TODO: read base part-by-part, not all at once, if scale > 2(?)
        image = self.base[src_loc]

        h, w = ((s.stop - s.start) for s in loc)
        return cv2.resize(image, (w, h), interpolation=cv2.INTER_AREA)


class Driver:
    @final
    @classmethod
    def register(cls, regex: str):
        """Registers type builder for extensions. Last call takes precedence"""
        REGISTRY.setdefault(re.compile(regex), []).append(cls)

    def __init__(self, path: str) -> None:
        raise NotImplementedError

    def __len__(self) -> int:
        """Count of indexed items"""
        return 0

    def __getitem__(self, index: int) -> Item:
        """Gives indexed item"""
        raise NotImplementedError

    def named_items(self) -> dict[str, Item]:
        keys = self.keys()
        return {k: self.get(k) for k in keys}

    def keys(self) -> list[str]:
        """List of names for named items"""
        return []

    def get(self, key: str) -> Item:
        """Gives named item"""
        raise NotImplementedError

    @property
    def bbox(self) -> tuple[slice, slice]:
        return slice(None), slice(None)



__all__ = ['env']

from pydantic import BaseSettings, ByteSize, parse_obj_as


class Env(BaseSettings):
    BIPL_DRIVERS: set[str] = {'gdal', 'tiff', 'openslide'}
    BIPL_CACHE: ByteSize = parse_obj_as(ByteSize, '10 MiB')


env = Env()



from collections.abc import Mapping
from itertools import zip_longest
from typing import Any

from lxml import etree
from lxml.etree import XMLParser, fromstring

# ------------------------- flat list of properties --------------------------


def unflatten(d: Mapping[str, str]) -> dict[str, Any]:
    if not d:
        return {}

    tab = {ord('['): '.', ord(']'): None}
    pairs = sorted((_make_key(k, tab), v) for k, v in d.items())
    r: dict = {}

    for parts, v in pairs:
        rr = r
        for curr, next_ in zip_longest(parts, parts[1:], fillvalue=None):
            # Put list if next is list idx, otherwise put dict
            default: list | dict | str
            default = v if next_ is None else (
                [] if isinstance(next_, int) else {})

            if isinstance(curr, int):  # last was list idx
                if not isinstance(rr, list):  # Ambigous structure
                    break
                if curr >= len(rr):
                    rr.extend(None for _ in range(len(rr), curr + 1))
                    rr[curr] = default
                rr = rr[curr]
            else:  # last was dict key
                if not isinstance(rr, dict):  # Ambigous structure
                    break
                rr = rr.setdefault(curr, default)
    return r


def _make_key(s: str, tab: Mapping[int, str | None]) -> tuple[str | int, ...]:
    parts = s.translate(tab).split('.')
    return tuple(int(word) if word.isdigit() else word for word in parts)


# ---------------------------- aperio description ----------------------------


def is_aperio(s: str) -> bool:
    return s.startswith('Aperio')


def parse_aperio_description(s: str) -> tuple[list[str], dict[str, str]]:
    header, *kv_pairs = s.split('|')

    head = [s.strip() for s in header.splitlines()]

    meta = {}
    for s in kv_pairs:
        tags = s.split('=', 1)
        if len(tags) == 2:
            meta[tags[0].strip()] = tags[1].strip()
        else:
            raise ValueError(f'Unparseable line in description: {s!r}')

    return head, meta


# ----------------------------- xml description ------------------------------


def parse_xml(s: str,
              /,
              *,
              group: str = 'property',
              name: str = 'name',
              value: str = 'value') -> dict[str, str]:
    t = fromstring(s, XMLParser(resolve_entities=False, no_network=True))

    # Remove a namespace URI in the element's name
    for elem in t.getiterator():
        if not isinstance(elem, etree._Comment | etree._ProcessingInstruction):
            elem.tag = etree.QName(elem).localname

    # Remove unused namespace declarations
    etree.cleanup_namespaces(t, top_nsmap=None, keep_ns_prefixes=None)

    return {e.find(name).text: e.find(value).text for e in t.iter(group)}


# ----------------------------------------------------------------------------


def _gdal_parse_description(meta: Mapping[str, str]) -> dict[str, Any]:
    desc = meta.get('TIFFTAG_IMAGEDESCRIPTION')
    if not desc:
        return {}
    if is_aperio(desc):
        raise ValueError('Aperio is not yet supported by GDAL driver')
    try:
        props = parse_xml(desc)
    except Exception:  # noqa: BLE001
        return {}
    return unflatten(props)


def gdal_parse_mpp(meta: Mapping) -> list[float]:
    props = _gdal_parse_description(meta)

    # TIFF tags
    tiff_res_tags = [f'TIFFTAG_{a}RESOLUTION' for a in 'XY']
    if res := [float(meta[trt]) for trt in tiff_res_tags if trt in meta]:
        if meta.get('TIFFTAG_RESOLUTIONUNIT') == '3 (pixels/cm)':
            return [10_000 / r for r in res]
        raise NotImplementedError

    # VIPS tags
    res = [float(props[tag]) for tag in ('xres', 'yres') if tag in props]
    if res:
        return [1_000 / r for r in res]

    # Openslide tags
    if osd := props.get('openslide'):
        mpp = [float(osd[tag]) for tag in ('mpp-x', 'mpp-y') if tag in osd]
        if mpp:
            return mpp

    # Aperio tags
    if mpp_ := props.get('MPP'):
        return [float(mpp_), float(mpp_)]

    return []



from ._mosaic import Mosaic, Tile, get_fusion
from ._util import get_trapz, normalize_loc, probs_to_rgb_heatmap

__all__ = [
    'Mosaic',
    'Tile',
    'get_fusion',
    'get_trapz',
    'normalize_loc',
    'probs_to_rgb_heatmap',
]



__all__ = ['get_trapz', 'normalize_loc', 'probs_to_rgb_heatmap']

import cv2
import numpy as np


def probs_to_rgb_heatmap(prob: np.ndarray) -> np.ndarray:
    """Converts probability array of (H W C) shape to (H W 3) RGB image"""
    if prob.ndim != 3:
        raise ValueError(f'prob should be 3d, got: {prob.shape}')
    h, w, c = prob.shape
    if not prob.size:
        return np.empty((h, w, 3), dtype='u1')

    hue = prob.argmax(-1).astype('u1')
    value = prob.max(-1)
    if value.dtype != 'u1':
        value *= 255
        value = value.round().astype('u1')

    hue_lut = np.zeros(256, 'u1')
    hue_lut[:c] = np.linspace(0, 127, c, endpoint=False, dtype='u1')
    hsv = cv2.merge([
        cv2.LUT(hue, hue_lut),
        np.broadcast_to(np.uint8(255), (h, w)),
        value,
    ])
    return cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB)


def get_trapz(step: int, overlap: int) -> np.ndarray:
    """Returns trapezional window kernel to apply stiching"""
    if overlap == 0:
        raise ValueError('not applicable for overlap 0')
    pad = np.linspace(0, 1, overlap + 2)[1:-1]  # strip initial 0 and final 1
    return np.r_[pad, np.ones(step - overlap), pad[::-1]].astype('f4')


def normalize_loc(loc: tuple[slice, ...] | slice,
                  shape: tuple[int, ...]) -> tuple[slice, ...]:
    """Ensures slices match shape and have non-none endpoints"""
    if isinstance(loc, slice):
        loc = loc,
    loc += (slice(None), ) * max(0, len(shape) - len(loc))
    if len(loc) != len(shape):
        raise ValueError(f'loc is too deep, got: {loc}')
    return *(slice(
        s.start if s.start is not None else 0,
        s.stop if s.stop is not None else axis_len,
        s.step if s.step is not None else 1,
    ) for s, axis_len in zip(loc, shape)),
